//==================================
// BELOW1MB.C - Matt Pietrek 1993
//==================================
#include <windows.h>
#include <toolhelp.h>
#include "below1mb.h"

HANDLE HInstance;
HWND HWndFreeLb;
HWND HWndUsedLb;
HWND HWndStatus;

DWORD EndOfLastFixedBlock;  // linear address of end of last fixed block
DWORD FreeBlocksTotalSize;  // Byte count of free regions < 1Mb

//
// Given a global handle, get the module name of its owner
//
void GetModuleNameFromHandle(HANDLE handle, char *owner)
{
    MODULEENTRY me;
    TASKENTRY   te;

    te.dwSize = sizeof(te);
    if ( TaskFindHandle(&te, handle) )  // First see if a task owns it
    {
        lstrcpy(owner, te.szModule);
        return;
    }

    me.dwSize = sizeof(me);
    if (ModuleFindHandle(&me, handle))  // Nope?  Try a module owner
    {
        lstrcpy(owner, me.szModule);
        return;
    }
    
    owner[0] = 0;   // No owner found.  Return empty string
}

// More descriptive strings for the TOOLHELP GT_xxx #define's
char *BlockTypes[] = {
"ALLOC", "DGROUP", "DATA", "CODE", "TDB", "RESOURCE", "MODULE",
"FREE", "INTERNAL", "SENTINEL", "BURGERMASTER" };

//
// Given a new memory block, see if is fixed or locked.  If so, it
// may indicate the end of a free region.  If so, add the info
// about the free region to the left listbox.  Then add the block's
// info to the right listbox.
//
void AddHeapEntryToListbox(GLOBALENTRY *ge)
{
    char buffer[128];
    char owner[32];

    if ( ge->wType != GT_FREE )     // If not a free block...
    {
        if ( !ge->hBlock )          // Throw out sentinel blocks
            return;                 // with no handles

        //
        // If the block can't be moved (i.e., it's fixed or
        // has a lock on it, see if there was a free region
        // beforehand.
        //
        if ( (ge->hBlock & 1) || ge->wcLock || ge->wcPageLock )
        {
            //
            // Was there a gap between the end of the last fixed block
            // and the start of this one?  If so, there was a free region.
            // Add the relevant information to the left listbox
            //
            if ( ge->dwAddress > EndOfLastFixedBlock )
            {
                DWORD freeBlockLen = ge->dwAddress-EndOfLastFixedBlock;
                wsprintf(buffer, "%08lX  %04lX bytes",
                    EndOfLastFixedBlock, freeBlockLen);
                SendMessage(HWndFreeLb, LB_ADDSTRING, 0,
                    (LPARAM)(LPSTR)buffer);
                FreeBlocksTotalSize += freeBlockLen;
            }
                
            EndOfLastFixedBlock = ge->dwAddress + ge->dwBlockSize;

            //
            // Add this blocks information to the right listbox
            //
            GetModuleNameFromHandle(ge->hOwner, owner);
            wsprintf(buffer,"%08lX  %04lX bytes (%s %s %s)",ge->dwAddress,
                    ge->dwBlockSize, (LPSTR)owner,
                    (LPSTR)BlockTypes[ge->wType], 
                    (LPSTR)((ge->hBlock & 1) ? "Fixed" : "Locked"));
            SendMessage(HWndUsedLb,LB_ADDSTRING,0,(LPARAM)(LPSTR)buffer);
        }
    }   
}

//
// Walk the heap below one megabyte.  The command parameter indicates
// whether the listboxes should be updated, or if the memory allocated
// by the "Allocate All" button should be freed.
//
void WalkHeap(unsigned command)
{
    GLOBALENTRY ge;
    BOOL moreToGo;
    char buffer[128];
    DWORD lastBelowOneMeg = 0;

    ge.dwSize = sizeof(ge);
    moreToGo = GlobalFirst( &ge, GLOBAL_ALL );
    EndOfLastFixedBlock = ge.dwAddress;
    FreeBlocksTotalSize = 0;
    
    while ( moreToGo )
    {
        WORD selectorToFree = 0;    // For WALK_HEAP_FREE_ALL
        
        // Break out of the loop when we encounter a block above 1 Meg
        if ( ge.dwAddress>=0x100000LU )
        {
            // We need to fake a FIXED block at the end of the heap
            // to finish off any possible free regions
            if ( command==WALK_HEAP_FILL_LISTBOXES )
            {
                ge.dwAddress = lastBelowOneMeg;
                ge.hBlock = 0xFFFF;     // A pseudo FIXED handle
                ge.wType = GT_SENTINEL;
                AddHeapEntryToListbox( &ge );
            }
            break;  // Break out of loop
        }
        
        if ( command == WALK_HEAP_FILL_LISTBOXES )
            AddHeapEntryToListbox( &ge );
        else    // command == WALK_HEAP_FREE_ALL
        {
            // If the blocks looks like a previous GlobalDosAlloc()
            // done by us, remember the handle for subsequent freeing.
            // If we free it here, the heap walk gets screwed up.
            if ( ge.hOwner == GetCurrentTask() )
            {
                if ((ge.wType == GT_UNKNOWN) && (ge.dwBlockSize==0x200))
                    selectorToFree = ge.hBlock;
            }
        }
        
        lastBelowOneMeg = ge.dwAddress + ge.dwBlockSize;
        moreToGo = GlobalNext(&ge, GLOBAL_ALL);
        if ( selectorToFree )   // Now we can safely free the block
            GlobalDosFree(selectorToFree);
    }

    // update the Total Free Space field
    wsprintf(buffer, "Total Free Space: %04lXh (%lu) bytes",
        FreeBlocksTotalSize, FreeBlocksTotalSize);
    SetWindowText(HWndStatus, buffer);
}

//
// Handle the pressing of any of the 3 buttons.
//
void Handle_WM_COMMAND(HWND hWndDlg, WPARAM wParam, LPARAM lParam)
{
    BOOL updateDisplay = FALSE;
    HCURSOR oldCursor = 0;      // Some operations take awhile...
    
    switch ( wParam )
    {   
        case IDC_BUTTON_ALLOC_ALL:
            oldCursor = SetCursor( LoadCursor(0, IDC_WAIT) );
            while (GlobalDosAlloc(0x200)) ; // Allocate it all!
            updateDisplay = TRUE;
            break;
        case IDC_BUTTON_FREE_ALL:
            oldCursor = SetCursor( LoadCursor(0, IDC_WAIT) );
            WalkHeap(WALK_HEAP_FREE_ALL);
            updateDisplay = TRUE;
            break;
        case IDC_BUTTON_REWALK:
            oldCursor = SetCursor( LoadCursor(0, IDC_WAIT) );
            updateDisplay = TRUE;
            break;
    }

    //
    // Clear out the listboxes, set immediate updating to false,
    // and walk the heap.  Afterwards, turn listbox updating back on.
    //
    if ( updateDisplay )
    {
        SendMessage(HWndFreeLb, LB_RESETCONTENT, 0, 0);
        SendMessage(HWndFreeLb, WM_SETREDRAW, FALSE, NULL);
        SendMessage(HWndUsedLb, LB_RESETCONTENT, 0, 0);         
        SendMessage(HWndUsedLb, WM_SETREDRAW, FALSE, NULL);
        WalkHeap( WALK_HEAP_FILL_LISTBOXES );
        SendMessage(HWndFreeLb, WM_SETREDRAW, TRUE, NULL);
        SendMessage(HWndUsedLb, WM_SETREDRAW, TRUE, NULL);
    }
    if ( oldCursor )            // Switch back to the normal cursor
        SetCursor(oldCursor);
}

void Handle_WM_INITDIALOG(HWND hWndDlg)
{
    // Get the HWND's of the commonly modified controls
    HWndFreeLb = GetDlgItem(hWndDlg, IDC_LISTBOX_FREE);
    HWndUsedLb = GetDlgItem(hWndDlg, IDC_LISTBOX_USED);
    HWndStatus = GetDlgItem(hWndDlg, IDC_TEXT_STATUS);

    // set the icon for when we're minimized
    SetClassWord(hWndDlg, GCW_HICON, LoadIcon(HInstance, "Below1MbIcon"));

    // Fake a "Re-Walk" button event
    Handle_WM_COMMAND(hWndDlg, IDC_BUTTON_REWALK, 0); 
}

//
// Dialog proc for the main dialog
//
BOOL CALLBACK _export Below1MbDlgProc(HWND hWndDlg, UINT msg,
                                           WPARAM wParam, LPARAM lParam)
{
    switch ( msg )
    {
        case WM_COMMAND:
            Handle_WM_COMMAND(hWndDlg, wParam, lParam); return TRUE; 
        case WM_INITDIALOG:
            Handle_WM_INITDIALOG(hWndDlg); return TRUE;
        case WM_CLOSE:
            EndDialog(hWndDlg, 0); return FALSE;
    }    
    return FALSE;
}

int PASCAL WinMain( HANDLE hInstance, HANDLE hPrevInstance,
                    LPSTR lpszCmdLine, int nCmdShow )
{
    if ( hPrevInstance )
        return 0;

    HInstance = hInstance;
    DialogBox(hInstance, "Below1MbDlg", 0, Below1MbDlgProc);
    return 0;
}
